#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Configuration example for Executer-srv version >= 2.0.0

!!! USE ONLY AS TEMPLATE !!!
"""
### Common section 
import PyTango
import time
# dynamic attribute class definition; not needed but easier to handle attr list
# import from colibry (preferred solution) or defined here 
from colibry.ds.executer import DynAttribute
'''class DynAttribute:
    def __init__(self, tango_attr_name, data_type):
        self.datatype = data_type # use native PyTango types
        self.handler = "self"
        self.tango_name = tango_attr_name
        self.name = tango_attr_name
        self.description = ""
        self.dimension = 1
        self.memorized = False
        self.default = None # use native type
        self.permission = PyTango.AttrWriteType.READ_WRITE
        self.read_function = None
        self.write_function = None
'''
### BL specific - dynamic attributes
# tango device names
RNM_EXPORTER_DEV = 'timex/daq/rnmexport'
SL_AMP1_TRIGGER_DEV = 'f/timing/rttrigger_sl.02' # this is SL in FEL1
SL_AMP2_TRIGGER_DEV = 'f/timing/rttrigger_sl.03' # this is SL in FEL2
SLU_SHUTTER_DEV = 'timex/laser/rt_shutter.01'

# library imports and global variables
import colibry.fermi as cbf
proxyManager = cbf.ProxyManager()
fermiInfo = cbf.FermiInfo(proxyManager)
fermiInfo.BunchnumberProxy = RNM_EXPORTER_DEV

# Instantiate dynamic attributes dictionary
# the Executer tango ds will expect this function in the conf script
def createDynAttrsDict():
    dyn_attrs = {}
    # manually created list of attributes with corresponding type
    attr_list = [("FilesToAcquire",PyTango.ArgType.DevShort),
             ("FileSize",PyTango.ArgType.DevShort),
             ("Project",PyTango.ArgType.DevString),
             ("Experiment",PyTango.ArgType.DevString),
             ("Dataset",PyTango.ArgType.DevString),
             ("LogEntry",PyTango.ArgType.DevString),
             ("Simulation",PyTango.ArgType.DevBoolean),
             ("Decimation_Frequency",PyTango.ArgType.DevDouble),
             ("Enable_SLU_Decimation",PyTango.ArgType.DevBoolean),
             ]
        
    # initialize dictionary with default information
    for attr in attr_list:
        dyn_attrs[attr[0]] = DynAttribute(attr[0],attr[1])
        
    # add further info specifically
    dyn_attrs["FilesToAcquire"].default = 1 
    dyn_attrs["FileSize"].default = 1
    dyn_attrs["Project"].memorized = True
    dyn_attrs["Experiment"].memorized = True
    dyn_attrs["Dataset"].memorized = True
    dyn_attrs["Decimation_Frequency"].read_function = "read_decimation_freq"
    dyn_attrs["Decimation_Frequency"].permission = PyTango.AttrWriteType.READ
    dyn_attrs["Enable_SLU_Decimation"].write_function = "toggle_slu_decimation"
    
    return dyn_attrs # the function will return the dictionary to the executer

### BL specific - Generic functions

# run a FEL sequence + one or more rttrigger devices
def create_sequence(good_shots, blank_shots, slu = False, shutter = None):
    global proxyManager
    global fermiInfo
    active_fel = fermiInfo.Active()
    felSeq = cbf.Sequence(proxyManager)
    felSeq.InfiniteCycle = True
    
    felSeq.CreateSequence("dir_seq",blank_shots,good_shots)
    # if the inverted seq for devices is required
    felSeq.CreateSequence("inv_seq",good_shots,blank_shots) 
    
    if shutter:
        felSeq.AddDevice(shutter, "dir_seq")
    else:
        if slu:
            dev1 = SL_AMP2_TRIGGER_DEV
            dev2 = SL_AMP1_TRIGGER_DEV
        else:
            dev1 = SL_AMP1_TRIGGER_DEV
            dev2 = SL_AMP2_TRIGGER_DEV
        
        if active_fel == "FEL1":
            felSeq.AddDevice(dev1, "dir_seq")
        elif active_fel == "FEL2":
            felSeq.AddDevice(dev2, "dir_seq")
    
    felSeq.SetupDevices()
    
    return felSeq

### BL specific - read/write functions for dynamic attributes

#-----------------------------------------------------------------------------------
#    toggle_SLU_decimation
#-----------------------------------------------------------------------------------
def toggle_slu_decimation(exec_obj,value_in):
    global proxyManager
    global fermiInfo
    good_shots = exec_obj.myattrs["Fermi_Shots"]
    blank_shots = exec_obj.myattrs["Fermi_Blanks"]

    # check for reasonable values
    if good_shots < 1 or blank_shots > 50 or good_shots > 50:
        return False;    
    
    sluSeq = create_sequence(good_shots, blank_shots, slu = True, shutter = SLU_SHUTTER_DEV)
    
    if value_in:
        # start always with open shutter
        try:
            Shutter_Proxy = proxyManager.AddProxy(SLU_SHUTTER_DEV)
            Shutter_Proxy.Open() # force open
        except:
            return False
        sluSeq.Start()
    else:
        sluSeq.Stop()
        
        # force shutter open
        time.sleep(0.5)
        try:
            Shutter_Proxy = proxyManager.AddProxy(SLU_SHUTTER_DEV)
            Shutter_Proxy.Open()
        except:
            pass      
    
    return value_in

#-----------------------------------------------------------------------------------
#    read_decimation_freq : calculated decimation frequency
#-----------------------------------------------------------------------------------
def read_decimation_freq(exec_obj):
    global fermiInfo
    good_shots = exec_obj.myattrs["Fermi_Shots"]
    blank_shots = exec_obj.myattrs["Fermi_Blanks"]
    fel_freq = fermiInfo.Frequency()
    frequency = fel_freq * (good_shots*1.)/(good_shots+blank_shots)
    return frequency