#!/usr/bin/env python3
#    "$Name:  $";
#    "$Header:  $";
#=============================================================================
#
# file :        Executer.py
#
# description : Python source for the Executer and its commands. 
#                The class is derived from Device. It represents the
#                CORBA servant object which will be accessed from the
#                network. All commands which can be executed on the
#                Executer are implemented in this file.
#
# project :     TANGO Device Server
#
# $Author:  $
#
# $Revision:  $
#
# $Log:  $
#
# copyleft :    European Synchrotron Radiation Facility
#               BP 220, Grenoble 38043
#               FRANCE
#
#=============================================================================
#          This file is generated by POGO
#    (Program Obviously used to Generate tango Object)
#
#         (c) - Software Engineering Group - ESRF
#=============================================================================
#
"""
Executer Server. 



"""
import sys
import os
sys.path.insert(0, os.path.abspath('.'))
if sys.version_info >= (3,):
    from collections import UserDict
else:
    from UserDict import UserDict
import PyTango
import time
import numpy as np
import subprocess
import logging
import logging.handlers
import importlib

import signal

def fromPyTango2NumpyType(pytango_type):
    """ Return the numpy type based on the PyTango type
    
    usage: 
        attr=PyTango.AttributeProxy('')
        sample=attr.read()
        fromPyTango2NumpyType(sample.type)
    """    
    data_type = np.int32
    if (pytango_type==PyTango.CmdArgType.DevUChar):
        data_type=np.uint8
    elif(pytango_type==PyTango.CmdArgType.DevUShort):
        data_type = np.uint16
    elif(pytango_type==PyTango.CmdArgType.DevULong):
        data_type = np.uint32
    elif(pytango_type==PyTango.CmdArgType.DevShort):
        data_type = np.int16
    elif(pytango_type==PyTango.CmdArgType.DevLong):
        data_type = np.int32
    elif(pytango_type==PyTango.CmdArgType.DevFloat):
        data_type = np.float32
    elif(pytango_type==PyTango.CmdArgType.DevDouble):
        data_type = np.float32 #FIXME version 1.2.1 does not support float64
    elif(pytango_type==PyTango.CmdArgType.DevBoolean):
        data_type = np.bool
    return data_type


def fromNumpyType2PyTango(numpy_type):
    """ Return the PyTango type based on numpy type
    
    usage: 
        attr = numpy.ones((2,2))
        fromNumpyType2PyTango(attr.type)                
    """        
    data_type = PyTango.ArgType.DevDouble
    if numpy_type == np.uint8:
        data_type = PyTango.ArgType.DevUChar
    elif numpy_type == np.uint16:
        data_type = PyTango.ArgType.DevUShort
    elif numpy_type == np.uint32:
        data_type = PyTango.ArgType.DevULong
    elif numpy_type == np.uint64:
        data_type = PyTango.ArgType.DevULong
    elif numpy_type == np.int16:
        data_type = PyTango.ArgType.DevShort
    elif numpy_type == np.int32:
        data_type = PyTango.ArgType.DevLong
    elif numpy_type == np.int64:
        data_type = PyTango.ArgType.DevLong
    elif numpy_type == np.float32:
        data_type = PyTango.ArgType.DevDouble
    elif numpy_type == np.bool:
        data_type = PyTango.ArgType.DevBoolean
    return data_type

def processInputStringArray(inp_array):
    """ Process an input array and return the values for the creation of
    PyTango Attribute.
    
    The input array must have: 
        name = The name of the new PyTango Attribute.
        data_type = string DOUBLE,SHORT or LONG strings
        optionals:
        [X size] = number of data for 1 and 2 d arrays
        [Y size] = number of data for 2 d arrays        
    
    it will return: 
        (name,PyTangoType,[int,[int]])
    """
    name = inp_array[0].lower()
    opt_type = inp_array[1].upper()
    if opt_type == "DOUBLE":
        typ = PyTango.ArgType.DevDouble
    elif opt_type == "SHORT":
        typ = PyTango.ArgType.DevShort
    elif opt_type == "LONG":
        typ = PyTango.ArgType.DevLong
    elif opt_type == "USHORT":
        typ = PyTango.ArgType.DevUShort
    elif opt_type == "ULONG":
        typ = PyTango.ArgType.DevULong
    elif opt_type == "FLOAT":
        typ = PyTango.ArgType.DevFloat        
    elif opt_type == "UCHAR":
        typ = PyTango.ArgType.DevUChar
    else:
        PyTango.Except.throw_exception("BadInput",
        "BadInput: (name,data_type,[xsize,[ysize]]), received: " + 
        str(inp_array) + " name = " +str(name) + " type = " +opt_type,
        "processStrings")
    
    if (len(inp_array) > 3):
        y_size = int(inp_array[3])
        x_size = int(inp_array[2])
        return (name,typ,x_size,y_size)
    elif (len(inp_array)>2):
        x_size = int(inp_array[2])
        return (name,typ,x_size)
    else:
        return (name,typ)

#==================================================================
#   Executer Class Description:
#
#         Creation and management of dynamic attributes that will be used for 
#         pre and pros-processing of Fermi data.
#
#==================================================================
#     Device States Description:
#
#   DevState.ON :     Working
#   DevState.FAULT :
#==================================================================


class CaseLessDict(UserDict):
    def __init__(self):
        UserDict.__init__(self)
    def __setitem__(self,key,v):
        UserDict.__setitem__(self,key.lower(),v)
    def __getitem__(self,key):
        key_l = key.lower()
        return UserDict.__getitem__(self,key_l)


class Executer(PyTango.Device_4Impl):

#--------- Add you global variables here --------------------------

#------------------------------------------------------------------
#    Device constructor
#------------------------------------------------------------------
    def __init__(self,cl, name):
        PyTango.Device_4Impl.__init__(self,cl,name)
        Executer.init_device(self)


#------------------------------------------------------------------
#    Device destructor
#------------------------------------------------------------------
    def delete_device(self):
        #print "[Device delete_device method] for device",self.get_name()
        if self.my_logger:
            del self.my_logger
        if self.script_executor:
            self.Abort()

#------------------------------------------------------------------
#    Device initialization
#------------------------------------------------------------------
    def init_device(self):
        #print "In ", self.get_name(), "::init_device()"
        self.set_state(PyTango.DevState.ON)
        self.get_device_properties(self.get_device_class())
        self.myattrs = CaseLessDict()
        self.read_functions = CaseLessDict()
        self.write_functions = CaseLessDict()
        #: Handler for the scripts to execute
        self.script_executor = None
        self._sta = ""
        self._historylines = []
        self._method2execute = "start"
        self.InteractiveBox = []
        # import the configuration script as a module
        try:
            self.cfg_module_name = os.path.splitext(os.path.basename(self.ConfiguratorScript))[0]
            self.cfg_module_path = os.path.dirname(self.ConfiguratorScript)
            print(self.cfg_module_name,self.cfg_module_path)
            sys.path.append(self.cfg_module_path)
            self.cfg_module = importlib.import_module(self.cfg_module_name)
            print("imported cfg")
        except Exception as ex:
            self.cfg_module = None
            #self.warn_stream("Invalid cfg  script file\n")
            print("Invalid cfg  script file\n",ex)
            self.set_state(PyTango.DevState.OFF)
            return
        
        # import the executable methods script as a module
        try:
            self.exec_module_name = os.path.splitext(os.path.basename(self.ExecutableScript))[0]
            self.exec_module_path = os.path.dirname(self.ExecutableScript)
            print(self.exec_module_name,self.exec_module_path)
            sys.path.append(self.exec_module_path)
            self.exec_module = importlib.import_module(self.exec_module_name)
            print("imported exec")
        except Exception as ex:
            sys.stderr.write(str(ex)+"\n")
            self.exec_module = None
            self.set_state(PyTango.DevState.FAULT)
            self.set_status("Invalid executable script file\n")
            return

        try:
            loggername = self.get_name().replace("/","-")
            self.my_logger = logging.getLogger(loggername)
            self.my_logger.setLevel(logging.DEBUG)
            # Add the log message handler to the logger
            LOG_FILENAME = self.LoggingPath + "/" + loggername +".log"
            handler = logging.handlers.RotatingFileHandler(LOG_FILENAME, maxBytes=1000000, backupCount=5)
            self.my_logger.addHandler(handler)
        except:
            self.my_logger = None
            print("Unable to open log file", LOG_FILENAME)
        
        #link to the Tango DB
        self.db = PyTango.Database()
        
        dynattrs = self.cfg_module.createDynAttrsDict()
        
        # creation of dynamic attributes
        for attr in dynattrs:
            attr_obj = dynattrs[attr]
            if not (attr_obj.handler.lower() == self.get_name().lower()
                or attr_obj.handler.lower() == 'self'): 
                try:
                    #print 'creation of attribute proxy: ',  str(value.tango_name)
                    attr_obj.tango_name = attr_obj.handler.lower() + "/" + attr_obj.tango_name
                    attr_proxy = PyTango.AttributeProxy(str(attr_obj.tango_name))
                    attr_value = attr_proxy.read()
                    attr_config = attr_proxy.get_config()
                    attr = PyTango.Attr(str(attr_obj.name), attr_config.data_type, 
                        attr_config.writable)
                    self.add_attribute(attr, self.read_proxy, self.write_proxy)
                    self.myattrs[attr_obj.name] = attr_proxy		
                except PyTango.DevFailed as ex: 
                    self.warn_stream("%s" % ex)		 
                continue
            if attr_obj.name in self.myattrs:
                PyTango.Except.throw_exception("Exist",
                            "Attribute "+attr_obj.name+ " already defined",
                            "NewAttribute")
            
            # It's possible to define python methods for calculating the attribute value
            self.read_functions[attr_obj.name] = attr_obj.read_function
            self.write_functions[attr_obj.name] = attr_obj.write_function
            
            if attr_obj.dimension == 1:                
                attr = PyTango.Attr(str(attr_obj.name), attr_obj.datatype, attr_obj.permission)
                if attr_obj.memorized:
                    attr.set_memorized()
                self.add_attribute(attr, self.read_Scalar,self.write_Scalar)
                if attr_obj.datatype == PyTango.ArgType.DevString:
                    self.myattrs[attr_obj.name] = ""
                else:
                    self.myattrs[attr_obj.name] = 0
                if attr_obj.default:
                    # find a better way to handle this
                    # using original types ??
                    self.myattrs[attr_obj.name] = attr_obj.default
                if attr_obj.memorized:
                    dev_attr_props = self.db.get_device_attribute_property(self.get_name(),[attr_obj.name])
                    attr_prop = dev_attr_props[attr_obj.name]
                    for pr in attr_prop.keys():
                        if (pr == '__value') and (attr_obj.datatype == PyTango.ArgType.DevString):
                            self.myattrs[attr_obj.name] = attr_prop['__value'][0]
                        elif (pr == '__value') and (attr_obj.datatype == PyTango.ArgType.DevFloat):
                            self.myattrs[attr_obj.name] = float(attr_prop['__value'][0])
                        elif (pr == '__value') and (attr_obj.datatype == PyTango.ArgType.DevDouble):
                            self.myattrs[attr_obj.name] = float(attr_prop['__value'][0])
                        elif (pr == '__value') and (attr_obj.datatype == PyTango.ArgType.DevBoolean):
                            self.myattrs[attr_obj.name] = bool(attr_prop['__value'][0] == "True")
                        elif (pr == '__value'):
                            self.myattrs[attr_obj.name] = int(attr_prop['__value'][0])
            elif isinstance(attr_obj.dimension, tuple):
                sizx = attr_obj.dimension[0]
                sizy = attr_obj.dimension[1]
                attr = PyTango.ImageAttr(str(attr_obj.name), attr_obj.datatype, attr_obj.permission,sizx, sizy)
                self.add_attribute(attr, self.read_Image, self.write_Image)
                if attr_obj.default:
                    self.myattrs[attr_obj.name] = attr_obj.default
                elif attr_obj.datatype == PyTango.ArgType.DevString:
                    self.myattrs[attr_obj.name] = [[],[]]
                else:
                    self.myattrs[attr_obj.name] = np.zeros((sizy,sizx),fromPyTango2NumpyType(attr_obj.datatype))
            else :
                self.debug_stream("Creation of attribute %s, dimension = %s" % (attr_obj.name, str(attr_obj.dimension)))
                attr = PyTango.SpectrumAttr(str(attr_obj.name), attr_obj.datatype, attr_obj.permission, 
                                    attr_obj.dimension)
                self.add_attribute(attr, self.read_Spectrum, self.write_Spectrum)
                if attr_obj.default:
                    self.myattrs[attr_obj.name] = attr_obj.default
                elif attr_obj.datatype == PyTango.ArgType.DevString:
                    self.myattrs[attr_obj.name] = []
                else:
                    self.myattrs[attr_obj.name] = np.zeros(attr_obj.dimension, fromPyTango2NumpyType(attr_obj.datatype))
        self.UpdateHistoryLog("Device initialized")
        self.set_state(PyTango.DevState.OFF)
                
#------------------------------------------------------------------
#    Always excuted hook method
#------------------------------------------------------------------
    def always_executed_hook(self):
        # self.debug_stream("In %s::always_excuted_hook()"%self.get_name())    
        if self.script_executor:
            if self.script_executor.poll() is None:
                if self.get_state() != PyTango.DevState.ON:
                    self.set_state(PyTango.DevState.ON)
            else:
                # Show errors if necessary
                for line in self.script_executor.stderr:
                    line = line.strip()
                    if sys.version_info >= (3,):
                        if isinstance(line, bytes):
                            line = line.decode("utf-8")
                        elif isinstance(line, bytearray):
                            line = ">>> bytearray found in hook."
                    self._historylines.append(line)
                #
                self.set_state(PyTango.DevState.OFF)
                self.script_executor = None

#==================================================================
#
#    Executer read/write attribute methods
#
#==================================================================
#------------------------------------------------------------------
#    Read Attribute Hardware
#------------------------------------------------------------------
    def read_attr_hardware(self,data):
        # self.debug_stream("In %s::read_attr_hardware()" % self.get_name())
        pass

#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def read_Method2Execute(self,attr):
        """ :returns: The function to invoke inside the script."""
        attr.set_value(self._method2execute)
    
#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def write_Method2Execute(self,attr):
        """ Set the function to invoke inside the script."""
        aux = attr.get_write_value(PyTango.ExtractAs.Numpy)
        self._method2execute = aux

#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def read_ScriptStatus(self,attr):
        """ :returns: The function to invoke inside the script."""
        attr.set_value(self._sta)
    

#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def read_StepByStepLog(self,attr):
        """ :returns: The function to invoke inside the script."""
        _historylog = "\n".join(self._historylines)
        attr.set_value(_historylog)


#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def read_InteractiveBox(self,attr):
        """
        :returns: InteractiveBox content
        """
        attr.set_value(self.InteractiveBox)


#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def write_InteractiveBox(self,attr):
        """
        Write InteractiveBox content.
        """
        value_in = attr.get_write_value()
        please_throw = False
        if len(value_in) != 3:
            please_throw = True
        elif value_in[0].upper() not in ["IN","OUT"]:
            please_throw = True
        elif value_in[1].lower() not in ["bool","float","str","int"]:
            please_throw = True
        if please_throw:
            PyTango.Except.throw_exception("Wrong format","Correct format is : ['IN'/'OUT', 'bool'/'int'/'float'/'str', str(message)/str(value)]","write_InteractiveBox()")
        #
        value_in[0] = value_in[0].upper()
        value_in[1] = value_in[1].lower()
        if value_in[0] == "IN":
            try:
                dummy = eval(value_in[1])(value_in[2])
                if value_in[1] == "bool":
                    if value_in[2].lower() in ["0","false"]:
                        value_in[2] = "False"
                    elif value_in[2].lower() in ["1","true"]:
                        value_in[2] = "True"
                    else:
                        PyTango.Except.throw_exception("Wrong value type","Cast error: "+value_in[1]+"('"+value_in[2]+"')","write_InteractiveBox()")
            except:
                PyTango.Except.throw_exception("Wrong value type","Cast error: "+value_in[1]+"('"+value_in[2]+"')","write_InteractiveBox()")
        #
        self.InteractiveBox = value_in

#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def UpdateStatus(self, aux):
        """ Set the function to invoke inside the script."""           
        self._sta = aux
        #self.info_stream("UpdateStatus = %s " % self._sta)


#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def UpdateHistoryLog(self, aux):
        """ Set the function to invoke inside the script."""
        if sys.version_info >= (3,):
            if isinstance(aux, bytes):
                aux = aux.decode("utf-8")
            elif isinstance(aux, bytearray):
                aux = ">>> bytearray found"
        now = time.localtime(time.time())
        now_str = '%02d:%02d:%02d ' %(now.tm_hour,now.tm_min,now.tm_sec)
        now_str += aux
        self._historylines.append(now_str)
        self._historylines = self._historylines[-self.Max_StepByStepLog_Lines:]
        if self.my_logger:
            datetime_str = '%02d/%02d/%04d ' %(now.tm_mday,now.tm_mon,now.tm_year)
            self.my_logger.debug(datetime_str + now_str)

#------------------------------------------------------------------
#    Read Scalar attribute
#------------------------------------------------------------------
    def read_Scalar(self, attr):
        #print "In ", self.get_name(), "::read_Scalar()"
        #    Add your own code here
        data = self.read_dynamic_attribute(attr.get_name())
        #
        attr.set_value(data)


#------------------------------------------------------------------
#    Write Scalar attribute
#------------------------------------------------------------------
    def write_Scalar(self, attr):
        #print "In ", self.get_name(), "::write_Scalar()"
        #    Add your own code here
        value_in = attr.get_write_value(PyTango.ExtractAs.Numpy) 
        self.write_dynamic_attribute(attr.get_name(), value_in)

#------------------------------------------------------------------
#    Read Spectrum attribute
#------------------------------------------------------------------
    def read_Spectrum(self, attr):
        #print "In ", self.get_name(), "::read_Spectrum()"
        #    Add your own code here
        data = self.read_dynamic_attribute(attr.get_name())
        #
        attr.set_value(data,len(data))        



#------------------------------------------------------------------
#    Write Spectrum attribute
#------------------------------------------------------------------
    def write_Spectrum(self, attr):
        #print "In ", self.get_name(), "::write_Spectrum()"
        #    Add your own code here
        value_in = attr.get_write_value(PyTango.ExtractAs.Numpy) 
        #
        self.write_dynamic_attribute(attr.get_name(), value_in)


#------------------------------------------------------------------
#    Read Image attribute
#------------------------------------------------------------------
    def read_Image(self, attr):
        #print "In ", self.get_name(), "::read_Image()"
        #    Add your own code here
        data = self.read_dynamic_attribute(attr.get_name())
        #
        assert isinstance(data,np.ndarray)
        #find_zero = np.where(data == 0)
        #data[find_zero] = data.max()
        attr.set_value(data)

#------------------------------------------------------------------
#    Write Image attribute
#------------------------------------------------------------------
    def write_Image(self, attr):
        #print "In ", self.get_name(), "::write_Image()"
        #    Add your own code here
        value_in = attr.get_write_value(PyTango.ExtractAs.Numpy) 
        self.write_dynamic_attribute(attr.get_name(), value_in)

#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def read_proxy(self, attr):
        val = self.myattrs[attr.get_name()].read()
        attr.set_value(val.value,val.dim_x, val.dim_y)

#------------------------------------------------------------------
#    
#------------------------------------------------------------------
    def write_proxy(self, attr):
        val = attr.get_write_value(PyTango.ExtractAs.Numpy)
        self.myattrs[attr.get_name()].write(val)

#==================================================================
#
#    Executer command methods
#
#==================================================================

    def is_Start_allowed(self):
        return self.get_state() == PyTango.DevState.OFF
        
    def TriggerMethod(self,argument):
        """ Executer the Method2Execute from the ExecutableScript with argument as input"""
        exec_cmd = "import %s as t; t.%s('%s')" %(self.exec_module_name, self._method2execute,argument)
        inp_list = [self.PythonExecutable,'-c',exec_cmd]
        self.script_executor = subprocess.Popen(inp_list, cwd = self.exec_module_path, stderr=subprocess.PIPE)


    def Start(self):
        """ Execute the Method2Execute from the ExecutableScript"""       
        del self._historylines[:]
        self.InteractiveBox = []
        if hasattr(self.exec_module, self._method2execute):
            exec_cmd = "import %s as t; t.%s()" %(self.exec_module_name, self._method2execute)
            inp_list = [self.PythonExecutable, '-c', exec_cmd]
            self.script_executor = subprocess.Popen(inp_list, cwd = self.exec_module_path, stderr=subprocess.PIPE)
            self.UpdateHistoryLog("Script started")
        else:
            self.UpdateHistoryLog("*** ERROR")
            msg = "Invalid Script: " + self._method2execute
            self.UpdateHistoryLog(msg)

    
    def StartMethod(self, method_name):
        if self.get_state() != PyTango.DevState.OFF:
            PyTango.Except.throw_exception("Unable to execute","Script already running","StartMethod")
        aux = self._method2execute
        self._method2execute = method_name
        self.Start()
        self._method2execute = aux


    
    def is_Abort_allowed(self):
        return self.get_state() == PyTango.DevState.ON
        
    def Abort(self):
        """ Abort the execution of script"""
        #print 'signal'
        self.UpdateHistoryLog("Aborting Script...")
        if hasattr(self.script_executor, 'send_signal'):
            #Python > 2.5
            self.script_executor.send_signal(signal.SIGABRT)
        else:
            #Python 2.5
            os.kill(self.script_executor.pid, signal.SIGABRT) 
        #self.script_executor.wait()
        #self.set_state(PyTango.DevState.OFF)
        #self.script_executor = None
        
#------------------------------------------------------------------
#    NewScalar command:
#
#    Description: 
#    argin:  DevVarStringArray    Name - of the new scalar attribute\nDOUBLE, SHORT, LONG for the type of the attribute\n
#------------------------------------------------------------------
    def NewScalar(self, argin):
        #print "In ", self.get_name(), "::NewScalar()"
        #    Add your own code here
        (name,typ) = processInputStringArray(argin)
        attr = PyTango.Attr(name, typ, PyTango.AttrWriteType.READ_WRITE)
        if self.myattrs.has_key(name):
            PyTango.Except.throw_exception("Exist","Attribute "+name+ " already defined","NewAttribute")
        self.add_attribute(attr, self.read_Scalar, self.write_Scalar,None)
        self.myattrs[name] = 0


#------------------------------------------------------------------
#    NewSpectrum command:
#
#    Description: 
#    argin:  DevVarStringArray    Name - of the new scalar attribute\nDOUBLE, SHORT, LONG for the type of the attribute\nMaxDim of the spectrum
#------------------------------------------------------------------
    def NewSpectrum(self, argin):
        #print "In ", self.get_name(), "::NewSpectrum()"
        #    Add your own code here
        (name,typ,siz) = processInputStringArray(argin)
        attr = PyTango.SpectrumAttr(name, typ, PyTango.AttrWriteType.READ_WRITE, siz)
        if self.myattrs.has_key(name):
            PyTango.Except.throw_exception("Exist","Attribute "+name+ " already defined","NewAttribute")                                        
        self.add_attribute(attr, self.read_Spectrum, self.write_Spectrum,None)
        self.myattrs[name] = np.zeros(siz,fromPyTango2NumpyType(typ))                                        
                                                
#------------------------------------------------------------------
#    NewImage command:
#
#    Description: 
#    argin:  DevVarStringArray    Name - of the new scalar attribute\nDOUBLE, SHORT, LONG for the type of the attribute\nMaxDimX of the image\nMaxDimY of the image
#------------------------------------------------------------------
    def NewImage(self, argin):
        #print "In ", self.get_name(), "::NewImage()"
        #    Add your own code here
        (name,typ,sizx,sizy) = processInputStringArray(argin)
        attr = PyTango.ImageAttr(name, typ, PyTango.AttrWriteType.READ_WRITE, sizx, sizy)
        if self.myattrs.has_key(name):
            PyTango.Except.throw_exception("Exist","Attribute "+name+ " already defined","NewAttribute")                                        
        self.add_attribute(attr, self.read_Image, self.write_Image,None)
        self.myattrs[name] = np.zeros((sizy,sizx),fromPyTango2NumpyType(typ))
        #print      self.myattrs[name].dtype   

#------------------------------------------------------------------
#
#------------------------------------------------------------------
    def ClearAll(self):
        #print "In ", self.get_name(), "::ClearAll()"
        for k in self.myattrs.keys():
            #print ' attempt to remove key = ' ,k
            try:
                self.remove_attribute(k.lower())
            except :
                pass
        #
        self.myattrs = CaseLessDict()
        self.read_functions = CaseLessDict()
        self.write_functions = CaseLessDict()


#==================================================================
#
#    ExecuterClass class utilities
#
#==================================================================

#------------------------------------------------------------------
#
#------------------------------------------------------------------
    def read_dynamic_attribute(self, attr_name):
        if self.read_functions[attr_name]:
            read_method = getattr(self.cfg_module,self.read_functions[attr_name])
            num_of_args = read_method.__code__.co_argcount
            if num_of_args == 1:
                # If method allows an input parameter,
                # it's a reference to this class
                data = read_method(self)
            else:
                data = read_method()
        else:
            data = self.myattrs[attr_name]
        return data

#------------------------------------------------------------------
#
#------------------------------------------------------------------
    def write_dynamic_attribute(self, attr_name, value_in):
        if self.write_functions[attr_name]:
            write_method = getattr(self.cfg_module,self.write_functions[attr_name])
            num_of_args = write_method.__code__.co_argcount
            if num_of_args == 2:
                # If method allows two input parameter,
                # the first is a reference to this class
                value = write_method(self, value_in)
            else:
                value = write_method(value_in)
        else:
            value = value_in
        self.myattrs[attr_name] = value
        if type(value_in) not in (int,float,str,bool):
            # Only scalar values can be memorized
            return
        dict_in = {attr_name:{'__value':value_in}}
        self.db.put_device_attribute_property(self.get_name(),dict_in)

#==================================================================
#
#    ExecuterClass class definition
#
#==================================================================
class ExecuterClass(PyTango.DeviceClass):

    #    Class Properties
    class_property_list = {
        }


    #    Device Properties
    device_property_list = {
        'ConfiguratorScript':
            [PyTango.DevString,
            "Path to the python config file",
            [ "./py_config_cfg.py"] ],
        'XmlConfigurator':
            [PyTango.DevString,
            "Path to the xml config file",
            [ "./xml_config.xml"] ],
        'ExecutableScript':
            [PyTango.DevString,
            "Path to the python script that will be executed.",
            [ "./test/script_test.py"] ],
        'PythonExecutable':
            [PyTango.DevString,
            "Path to the python executable.",
            [ "python"] ],
        'LoggingPath':
            [PyTango.DevString,
            "Path to the logging directory.",
            [ "/tmp/log"] ],
        'Max_StepByStepLog_Lines':
            [PyTango.DevShort,
            "Maximum number of line for StepByStepLog attribute",
            [ 2000 ] ],
        }


    #    Command definitions
    cmd_list = {
        'NewScalar':
            [[PyTango.DevVarStringArray, "Name - of the new scalar attribute\nDOUBLE, SHORT, LONG for the type of the attribute\n"],
            [PyTango.DevVoid, ""]],
        'NewSpectrum':
            [[PyTango.DevVarStringArray, "Name - of the new scalar attribute\nDOUBLE, SHORT, LONG for the type of the attribute\nMaxDim of the spectrum"],
            [PyTango.DevVoid, ""]],
        'NewImage':
            [[PyTango.DevVarStringArray, "Name - of the new scalar attribute\nDOUBLE, SHORT, LONG for the type of the attribute\nMaxDimX of the image\nMaxDimY of the image"],
            [PyTango.DevVoid, ""]],
        'ClearAll':
            [[PyTango.DevVoid,""],
             [PyTango.DevVoid,""]],
        'Start':
            [[PyTango.DevVoid,""],
             [PyTango.DevVoid,""]],
        'Abort':
            [[PyTango.DevVoid,""],
             [PyTango.DevVoid,""]],
        'StartMethod':
            [[PyTango.DevString,"Execute the method from the script"],
             [PyTango.DevVoid,""]],
        'TriggerMethod':
            [[PyTango.DevString,"The input string value for the Method2Execute script"],
             [PyTango.DevVoid,""]],
        'UpdateStatus':
            [[PyTango.DevString,"Status of the script"],
             [PyTango.DevVoid,""]], 
        'UpdateHistoryLog':
            [[PyTango.DevString,"Step by Step script log message"],
             [PyTango.DevVoid,""]], 
        }


    #    Attribute definitions
    attr_list = {
        'Method2Execute':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description':"The function inside the script that will be invoked at start command",
                'Memorized': "true",
            } ],
        'ScriptStatus':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description':"script information",
            } ],            
        'StepByStepLog':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ],
            {
                'description':"Script step-by-step log messages",
            } ],
        'InteractiveBox':
            [[PyTango.DevString,
              PyTango.SPECTRUM,
              PyTango.READ_WRITE, 3],
              {
                  'description':'Used for interactive operations: ["IN"/"OUT", "bool"/"int"/"float"/"str", str(message)/str(value)]',
                  'Memorized':"false",
            }],	    
#        'Spectrum':
#            [[PyTango.DevDouble,
#            PyTango.SPECTRUM,
#            PyTango.READ_WRITE, 10000],
#            {
#                'description':"Example of spectrum attribute.",
#            } ],
#        'Image':
#            [[PyTango.DevLong,
#            PyTango.IMAGE,
#            PyTango.READ_WRITE, 3000, 3000],
#            {
#                'description':"Example of an image attribute",
#            } ],
        }


#------------------------------------------------------------------
#    ExecuterClass Constructor
#------------------------------------------------------------------
    def __init__(self, name):
        PyTango.DeviceClass.__init__(self, name)
        self.set_type(name);
        #print "In ExecuterClass  constructor"


#==================================================================
#
#    Executer class main method
#
#==================================================================
def main():
    try:
        py = PyTango.Util(sys.argv)
        py.add_TgClass(ExecuterClass,Executer,'Executer')

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed as ex:
        print('-------> Received a DevFailed exception:',ex)
    except Exception as ex:
        print('-------> An unforeseen exception occured....',ex)


if __name__ == '__main__':
    main()
